use {
    crate::{
        VM_Word, VM_Instruction,
        linker::LinkAddr
    },
    serde::{Serialize, Deserialize}
};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum AssemblyWord {
    Symbol( LinkAddr ),
    Lit( VM_Word )
}

impl AssemblyWord {
    pub fn inst(i: VM_Instruction) -> Self {
        AssemblyWord::Lit(i as VM_Word)
    }

    pub fn lit(w: VM_Word) -> Self {
        AssemblyWord::Lit(w)
    }

    pub fn symbol(name: &str) -> Self {
        AssemblyWord::Symbol(LinkAddr::Relative{ symbol: name.into(), offset: 0 })
    }
}

pub fn disassemble(bytecode: &Vec< AssemblyWord >) -> Vec< String > {
    let mut words = bytecode.iter();
    let mut result = Vec::new();

    let mut remaining_params = 0;

    while let Some(w) = words.next() {
        match w {
            AssemblyWord::Symbol( addr ) => {
                result.push( format!("{}", addr) );
                remaining_params -= 1;
            }
            AssemblyWord::Lit(w) => {
                if remaining_params == 0 {
                    if *w < 32 {
                        let inst : VM_Instruction = unsafe { std::mem::transmute(*w) };
                        result.push(format!("{:?}", inst));
                        remaining_params = inst.param_length();
                    } else {
                        result.push(format!("invalid opcode {}", w));
                    }
                } else {
                    remaining_params -= 1;
                    result.push(format!("{}", w));
                }
            }
        }
    }

    result
}

pub struct Assembler {
    words: Vec< AssemblyWord >,
}

impl Assembler {
    pub fn new() -> Self {
        Assembler {
            words: Vec::new()
        }
    }

    pub fn build(mut self) -> Vec< AssemblyWord > {
        self.words.push( AssemblyWord::inst(VM_Instruction::Ret) );
        self.words
    }

    pub fn join(mut self, mut other: Assembler) -> Self {
        self.words.append(&mut other.words);
        self
    }

    pub fn lit(mut self, w: VM_Word) -> Assembler {
        self.words.push( AssemblyWord::inst(VM_Instruction::Lit) );
        self.words.push( AssemblyWord::lit(w) );
        self
    }

    pub fn litf(mut self, wf: f64) -> Assembler {
        let w : VM_Word = unsafe{ std::mem::transmute(wf) };
        self.words.push( AssemblyWord::inst(VM_Instruction::Lit) );
        self.words.push( AssemblyWord::lit(w) );
        self
    }

    pub fn inst(mut self, i: VM_Instruction) -> Assembler {
        self.words.push( AssemblyWord::inst(i) );
        self
    }

    pub fn call_symbol(mut self, addr: LinkAddr) -> Assembler {
        self.words.push( AssemblyWord::inst(VM_Instruction::Call) );
        self.words.push( AssemblyWord::Symbol(addr) );
        self
    }

    pub fn ref_symbol(mut self, addr: LinkAddr) -> Assembler {
        self.words.push( AssemblyWord::inst(VM_Instruction::Lit) );
        self.words.push( AssemblyWord::Symbol(addr) );
        self
    }

    pub fn call(mut self, s: &str) -> Assembler {
        self.words.push( AssemblyWord::inst(VM_Instruction::Call) );
        self.words.push( AssemblyWord::symbol(s) );
        self
    }

    pub fn static_ref(mut self, s: &str) -> Assembler {
        self.words.push( AssemblyWord::inst(VM_Instruction::Lit) );
        self.words.push( AssemblyWord::symbol(s) );
        self
    }

    pub fn while_loop(mut self, mut condition: Assembler, mut body: Assembler) -> Assembler {
        let cond_len = condition.words.len();
        let body_len = body.words.len();
        self.words.append( &mut condition.words );

        // jump to end
        self.words.push( AssemblyWord::inst(VM_Instruction::Branch) );
        self.words.push( AssemblyWord::lit(body_len as VM_Word + 2) );

        self.words.append( &mut body.words );

        // jump back to condition
        self.words.push( AssemblyWord::inst(VM_Instruction::Jmp) );
        self.words.push( AssemblyWord::lit(-2-(body_len as VM_Word)-2-(cond_len as VM_Word)) ); 

        self
    }
 
    pub fn branch(mut self, mut if_branch: Assembler, mut else_branch: Assembler) -> Assembler {
        self.words.push(AssemblyWord::inst(VM_Instruction::Branch));
        self.words.push(AssemblyWord::lit(if_branch.words.len() as VM_Word + 2));

        self.words.append( &mut if_branch.words );
        self.words.push(AssemblyWord::inst(VM_Instruction::Jmp));
        self.words.push(AssemblyWord::lit(else_branch.words.len() as VM_Word));

        self.words.append( &mut else_branch.words );

        self
    }
}

