use {
    std::collections::HashMap,
    crate::{VM_Word, assembler::AssemblyWord},
    serde::{Serialize, Deserialize}
};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum LinkAddr {
    Absolute(VM_Word),
    Relative {
        symbol: String,
        offset: VM_Word
    }
}

impl std::fmt::Display for LinkAddr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        match self {
            LinkAddr::Absolute(addr) =>
                write!(f, "{:#08x}", addr),
            LinkAddr::Relative{ symbol, offset } =>
                if *offset == 0 {
                    write!(f, "{}", symbol)
                } else {
                    write!(f, "{}+{:x}", symbol, offset)
                }
        }
    }
}

#[derive(Serialize, Deserialize, Debug)]
struct ObjAddr {
    section_id: usize,
    offset: i64
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Section {
    pub addr: VM_Word,
    pub data: Vec< AssemblyWord >
}

pub struct Linker {
    pub sections: Vec<Section>,
    pub symbols: HashMap<String, ObjAddr>,

    // internal linker state
    pub next_addr: VM_Word
}

#[derive(Clone, Debug)]
pub enum LinkError {
    UnresolvedSymbol( String ),
    RelativeAddrOutOfBounds( String, VM_Word )
}

#[derive(Serialize, Deserialize, Debug)]
pub struct ObjectFile {
    pub symbols: Vec< (String, VM_Word) >,
    pub code: Vec< AssemblyWord >,
}

impl Linker {
    pub fn new() -> Self {
        Linker {
            sections: Vec::new(),
            symbols: HashMap::new(),
            next_addr: 0
        }
    }

    pub fn import(&mut self, source_name: &str, obj_file: ObjectFile) {
        let section_id = self.add_procedure(
            &source_name,
            obj_file.code
        );

        for (symbol, offset) in obj_file.symbols {
            self.symbols.insert(symbol, ObjAddr {
                section_id,
                offset
            });
        }
    }

    pub fn link_partial(self) -> Result<Vec<AssemblyWord>, LinkError> {
        let mut bytecode = vec![ AssemblyWord::Lit(0); self.next_addr as usize ];
        for Section{ mut addr, data } in self.sections.iter() {
            for w in data {
                bytecode[ addr as usize ] = self.resolve_link(w.clone())?;
                addr += 1;
            }
        }
        Ok(bytecode)
    }

    pub fn link_relative(self, symbol: &String) -> Result<Vec<AssemblyWord>, LinkError> {
        Ok(self.link_partial()?
            .into_iter()
            .map(|w| match w {
                AssemblyWord::Symbol(LinkAddr::Absolute(addr)) =>
                    AssemblyWord::Symbol(LinkAddr::Relative {
                        symbol: symbol.clone(),
                        offset: addr
                    }),
                w => w
            })
            .collect())
    }

    pub fn link_total(self) -> Result<Vec<VM_Word>, LinkError> {
        self.link_partial()?
            .into_iter()
            .map(
                |w| match w {
                    AssemblyWord::Lit(w) => Ok(w),
                    AssemblyWord::Symbol( LinkAddr::Absolute(w) ) => Ok(w),
                    AssemblyWord::Symbol( LinkAddr::Relative{ symbol, offset }) => {
                        Err(LinkError::UnresolvedSymbol(symbol))
                    }
                }
            )
            .collect()
    }

    pub fn get_link_addr(&self, name: &String) -> Option<VM_Word> {
        match self.symbols.get(name) {
            Some(ObjAddr{ section_id, offset }) =>{
                let section = self.sections.get(*section_id).unwrap();
                Some(section.addr + offset)
            },
            None => None
        }
    }

    pub fn resolve_link(&self, w: AssemblyWord) -> Result<AssemblyWord, LinkError> {
        match w {
            AssemblyWord::Symbol(addr) =>
                Ok(AssemblyWord::Symbol( self.resolve_link_addr(addr)? )),
            AssemblyWord::Lit(w) => Ok(AssemblyWord::Lit(w))
        }
    }

    pub fn get_symbol_range( &self, name: &str ) -> Option<(i64, usize)> {
        if let Some(ObjAddr{ section_id, offset }) = self.symbols.get(name) {
            if let Some(Section{ data, addr }) = self.sections.get(*section_id) {
                Some((*addr as i64 + *offset as i64, data.len()))
            } else {
                None
            }
        } else {
            None
        }
    }

    pub fn resolve_link_addr(&self, addr: LinkAddr) -> Result<LinkAddr, LinkError> {
        match addr {
            LinkAddr::Absolute(w) => Ok(LinkAddr::Absolute(w)),
            LinkAddr::Relative{ symbol, offset } => {
                if let Some((symbol_addr, len)) = self.get_symbol_range(&symbol) {
                    if offset < len as i64 && offset >= 0 {
                        Ok(LinkAddr::Absolute( symbol_addr + offset as i64 ))
                    } else {
                        eprintln!("symbol {} @ {}+{}", symbol, symbol_addr, len);
                        Err(LinkError::RelativeAddrOutOfBounds(symbol, offset))
                    }
                } else {
                    Ok(LinkAddr::Relative{ symbol, offset })
                }
            }
        }
    }

    pub fn add_static(
        &mut self,
        symbol: &str,
        data: Vec<VM_Word>,
    ) -> usize {
        let link_addr = self.next_addr;
        self.next_addr += data.len() as i64;

        let section_id = self.sections.len();

        self.sections.push(Section {
            addr: link_addr,
            data: data.into_iter().map(|w| AssemblyWord::Lit(w)).collect()
        });

        self.symbols.insert(
            symbol.into(),
            ObjAddr {
                section_id,
                offset: 0
            }
        );

        section_id
    }

    pub fn add_procedure_front(
        &mut self,
        symbol: &str,
        bytecode: Vec< AssemblyWord >
    ) -> usize {
        self.next_addr += bytecode.len() as i64;
        for section in self.sections.iter_mut() {
            section.addr += bytecode.len() as i64;
        }

        let section_id = self.sections.len();

        self.sections.push(Section {
            addr: 0,
            data: bytecode
        });

        self.symbols.insert(symbol.into(), ObjAddr{
            section_id,
            offset: 0
        });

        section_id
    }

    pub fn add_procedure(
        &mut self,
        symbol: &str,
        bytecode: Vec< AssemblyWord >
    ) -> usize {
        let link_addr = self.next_addr;
        self.next_addr += bytecode.len() as i64;

        
        let section_id = self.sections.len();

        self.sections.push(Section {
            addr: link_addr,
            data: bytecode
        });

        self.symbols.insert(symbol.into(), ObjAddr{
            section_id,
            offset: 0
        });

        section_id
    }
}

